#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

from typing import Optional, Tuple

import numpy as np
import pinocchio as pin
import upkie_description
from gymnasium import spaces

from upkie.utils.clamp import clamp_and_warn
from upkie.utils.exceptions import ModelError
from upkie.utils.pinocchio import (
    box_position_limits,
    box_torque_limits,
    box_velocity_limits,
)
from upkie.utils.robot_state import RobotState

from .upkie_base_env import UpkieBaseEnv


class UpkieServos(UpkieBaseEnv):
    """!
    Upkie with with action and observation for each servo.

    ### Action space

    The action space is a dictionary with one key for each servo:

    - ``left_hip``: Left hip joint (qdd100)
    - ``left_hip``: Left knee joint (qdd100)
    - ``left_hip``: Left wheel joint (mj5208)
    - ``right_hip``: Right hip joint (qdd100)
    - ``right_hip``: Right knee joint (qdd100)
    - ``right_hip``: Right wheel joint (mj5208)

    The value for each dictionary is a dictionary with the following keys:

    - ``position``: Joint angle in [rad] (NaN to disable) (required).
    - ``velocity``: Joint velocity in [rad] / [s] (required).
    - ``feedforward_torque``: Joint torque in [N] * [m].
    - ``kp_scale``: Scaling factor applied to the position feedback gain,
        between zero and one.
    - ``kd_scale``: Scaling factor applied to the velocity feedback gain,
        between zero and one.
    - ``maximum_torque``: Maximum joint torque (feedforward + feedback) during
        the whole actuation step, in [N] * [m].

    ### Observation space

    The observation space is a dictionary with one key for each servo. The
    value for each key is a dictionary with keys:

    - ``position``: Joint angle in [rad].
    - ``velocity``: Joint velocity in [rad] / [s].
    - ``torque``: Joint torque in [N] * [m].
    - ``temperature``: Servo temperature in degree Celsius.
    - ``voltage": Power bus voltage of the servo, in [V].

    As with all Upkie environments, full observations from the spine (detailed
    in @ref observations) are also available in the ``info`` dictionary
    returned by the reset and step functions.
    """

    ACTION_KEYS: Tuple[str, str, str, str, str, str] = (
        "position",
        "velocity",
        "feedforward_torque",
        "kp_scale",
        "kd_scale",
        "maximum_torque",
    )

    JOINT_NAMES: Tuple[str, str, str, str, str, str] = (
        "left_hip",
        "left_knee",
        "left_wheel",
        "right_hip",
        "right_knee",
        "right_wheel",
    )

    ## @var action_space
    ## Action space.
    action_space: spaces.box.Box

    ## @var observation_space
    ## Observation space.
    observation_space: spaces.box.Box

    ## @var robot
    ## Pinocchio robot wrapper.
    robot: pin.RobotWrapper

    ## @var version
    ## Environment version number.
    version = 4

    def __init__(
        self,
        fall_pitch: float = 1.0,
        frequency: float = 200.0,
        frequency_checks: bool = True,
        init_state: Optional[RobotState] = None,
        regulate_frequency: bool = True,
        shm_name: str = "/vulp",
        spine_config: Optional[dict] = None,
    ):
        r"""!
        Initialize environment.

        \param fall_pitch Fall pitch angle, in radians.
        \param frequency Regulated frequency of the control loop, in Hz.
        \param frequency_checks If `regulate_frequency` is set and this
            parameter is true (default), a warning is issued every time the
            control loop runs slower than the desired `frequency`. Set this
            parameter to false to disable these warnings.
        \param init_state Initial state of the robot, only used in simulation.
        \param regulate_frequency Enables loop frequency regulation.
        \param shm_name Name of shared-memory file.
        \param spine_config Additional spine configuration overriding the
            defaults from ``//config:spine.yaml``. The combined configuration
            dictionary is sent to the spine at every :func:`reset`.
        """
        super().__init__(
            fall_pitch=fall_pitch,
            frequency=frequency,
            frequency_checks=frequency_checks,
            init_state=init_state,
            regulate_frequency=regulate_frequency,
            shm_name=shm_name,
            spine_config=spine_config,
        )

        robot = upkie_description.load_in_pinocchio(root_joint=None)
        model = robot.model
        q_min, q_max = box_position_limits(model)
        v_max = box_velocity_limits(model)
        tau_max = box_torque_limits(model)
        joint_names = list(model.names)[1:]
        if set(joint_names) != set(self.JOINT_NAMES):
            raise ModelError(
                "Upkie joints don't match:"
                f"expected {self.JOINT_NAMES} "
                f"got {joint_names}"
            )

        action_space = {}
        neutral_action = {}
        max_action = {}
        min_action = {}
        servo_space = {}
        for name in joint_names:
            joint = model.joints[model.getJointId(name)]
            action_space[name] = spaces.Dict(
                {
                    "position": spaces.Box(
                        low=q_min[joint.idx_q],
                        high=q_max[joint.idx_q],
                        shape=(1,),
                        dtype=float,
                    ),
                    "velocity": spaces.Box(
                        low=-v_max[joint.idx_v],
                        high=+v_max[joint.idx_v],
                        shape=(1,),
                        dtype=float,
                    ),
                    "feedforward_torque": spaces.Box(
                        low=-tau_max[joint.idx_v],
                        high=+tau_max[joint.idx_v],
                        shape=(1,),
                        dtype=float,
                    ),
                    "kp_scale": spaces.Box(
                        low=0.0,
                        high=1.0,
                        shape=(1,),
                        dtype=float,
                    ),
                    "kd_scale": spaces.Box(
                        low=0.0,
                        high=1.0,
                        shape=(1,),
                        dtype=float,
                    ),
                    "maximum_torque": spaces.Box(
                        low=0.0,
                        high=tau_max[joint.idx_v],
                        shape=(1,),
                        dtype=float,
                    ),
                }
            )
            servo_space[name] = spaces.Dict(
                {
                    "position": spaces.Box(
                        low=q_min[joint.idx_q],
                        high=q_max[joint.idx_q],
                        shape=(1,),
                        dtype=float,
                    ),
                    "velocity": spaces.Box(
                        low=-v_max[joint.idx_v],
                        high=+v_max[joint.idx_v],
                        shape=(1,),
                        dtype=float,
                    ),
                    "torque": spaces.Box(
                        low=-tau_max[joint.idx_v],
                        high=+tau_max[joint.idx_v],
                        shape=(1,),
                        dtype=float,
                    ),
                    "temperature": spaces.Box(
                        low=0.0,
                        high=100.0,
                        shape=(1,),
                        dtype=float,
                    ),
                    "voltage": spaces.Box(
                        low=10.0,  # moteus min 10 V
                        high=44.0,  # moteus max 44 V
                        shape=(1,),
                        dtype=float,
                    ),
                }
            )
            neutral_action[name] = {
                "position": np.nan,
                "velocity": 0.0,
                "feedforward_torque": 0.0,
                "kp_scale": 1.0,
                "kd_scale": 1.0,
                "maximum_torque": tau_max[joint.idx_v],
            }
            max_action[name] = {
                "position": q_max[joint.idx_q],
                "velocity": v_max[joint.idx_v],
                "feedforward_torque": tau_max[joint.idx_v],
                "kp_scale": 1.0,
                "kd_scale": 1.0,
                "maximum_torque": tau_max[joint.idx_v],
            }
            min_action[name] = {
                "position": q_min[joint.idx_q],
                "velocity": -v_max[joint.idx_v],
                "feedforward_torque": -tau_max[joint.idx_v],
                "kp_scale": 0.0,
                "kd_scale": 0.0,
                "maximum_torque": 0.0,
            }

        # gymnasium.Env: action_space
        self.action_space = spaces.Dict(action_space)

        # gymnasium.Env: observation_space
        self.observation_space = spaces.Dict(servo_space)

        # Class members
        self.__neutral_action = neutral_action
        self.__max_action = max_action
        self.__min_action = min_action
        self.robot = robot

    def get_neutral_action(self) -> dict:
        """!
        Get the neutral action where servos don't move.

        @returns Neutral action where servos don't move.
        """
        return self.__neutral_action.copy()

    def get_env_observation(self, spine_observation: dict):
        r"""!
        Extract environment observation from spine observation dictionary.

        \param spine_observation Full observation dictionary from the spine.
        \return Environment observation.
        """
        # If creating a new object turns out to be too slow we can switch to
        # updating in-place.
        return {
            joint: {
                key: np.array(
                    [spine_observation["servo"][joint][key]],
                    dtype=float,
                )
                for key in self.observation_space[joint]
            }
            for joint in self.JOINT_NAMES
        }

    def get_spine_action(self, env_action: dict) -> dict:
        r"""!
        Convert environment action to a spine action dictionary.

        \param env_action Environment action.
        \return Spine action dictionary.
        """
        spine_action = {"servo": {}}
        for joint in self.JOINT_NAMES:
            servo_action = {}
            for key in self.ACTION_KEYS:
                action = (
                    env_action[joint][key]
                    if key in env_action[joint]
                    else self.__neutral_action[joint][key]
                )
                servo_action[key] = clamp_and_warn(
                    action,
                    self.__min_action[joint][key],
                    self.__max_action[joint][key],
                    label=f"{joint}: {key}",
                )
            spine_action["servo"][joint] = servo_action
        return spine_action

    def get_reward(self, observation: dict, action: dict) -> float:
        r"""!
        Get reward from observation and action.

        \param observation Environment observation.
        \param action Environment action.
        \return Reward.
        """
        return 1.0



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

import os

import yaml

PATH = os.path.abspath(os.path.dirname(__file__))

SPINE_CONFIG = None

with open(PATH + "/spine.yaml", "r") as fh:
    SPINE_CONFIG = yaml.safe_load(fh)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria


class MockSpine:
    def __init__(self):
        self.observation = {
            "base_orientation": {
                "pitch": 0.1,
                "angular_velocity": [-2e-3, 3e4, 1e-8],
                "linear_velocity": [1e3, 2e2, 3e1],
            },
            "imu": {
                "orientation": [1.0, 0.0, 0.0, 0.0],
                "angular_velocity": [0.0, 0.0, 0.0],
                "linear_acceleration": [0.0, 0.0, 0.0],
            },
            "number": 0,
            "servo": {
                f"{side}_{joint}": {
                    "position": 0.0,
                    "velocity": 0.0,
                    "torque": 0.0,
                    "temperature": 42.0,
                    "voltage": 18.0,
                }
                for side in ("left", "right")
                for joint in ("hip", "knee", "wheel")
            },
            "wheel_odometry": {
                "position": 0.0,
                "velocity": 0.0,
            },
        }

    def start(self, config: dict) -> None:
        pass

    def stop(self) -> None:
        pass

    def set_action(self, action) -> None:
        self.action = action

    def get_observation(self) -> dict:
        self.observation["number"] += 1
        return self.observation



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

import logging

import gymnasium as gym

from .upkie_base_env import UpkieBaseEnv

__all__ = [
    "UpkieBaseEnv",
    "register",
]

__envs__ = {}

try:
    from .upkie_ground_velocity import UpkieGroundVelocity

    __all__.append("UpkieGroundVelocity")
    __envs__["UpkieGroundVelocity"] = UpkieGroundVelocity
except ImportError as exn:
    __envs__["UpkieGroundVelocity"] = exn

try:
    from .upkie_servos import UpkieServos

    __all__.append("UpkieServos")
    __envs__["UpkieServos"] = UpkieServos
except ImportError as exn:
    __envs__["UpkieServos"] = exn


def register() -> None:
    """!
    Register Upkie environments with Gymnasium.
    """
    for env_name, env_class in __envs__.items():
        if isinstance(env_class, ModuleNotFoundError):
            import_error = str(env_class)
            logging.warning(
                "Cannot register %s due to missing dependency: %s",
                env_name,
                import_error,
            )
        else:  # valid gym.Env subclass
            gym.envs.registration.register(
                id=f"{env_name}-v{env_class.version}",
                entry_point=f"upkie.envs:{env_name}",
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

from typing import Tuple

import numpy as np
import pinocchio as pin
from numpy.typing import NDArray


def box_position_limits(
    model: pin.Model,
) -> Tuple[NDArray[float], NDArray[float]]:
    r"""!
    Compute position limits in box format:

    \f[
    q_{min} \leq q \leq q_{max}
    \f]

    \param model Pinocchio model.
    @returns Tuple ``(q_min, q_max)`` of lower and upper position limits, with
    -infinity and +infinity where there is no limit.
    """
    no_position_limit = np.logical_or(
        model.upperPositionLimit > 1e20,
        model.upperPositionLimit < model.lowerPositionLimit + 1e-10,
    )
    q_min = model.lowerPositionLimit.copy()
    q_max = model.upperPositionLimit.copy()
    q_min[no_position_limit] = -np.inf
    q_max[no_position_limit] = +np.inf
    return q_min, q_max


def box_velocity_limits(
    model: pin.Model,
) -> Tuple[NDArray[float], NDArray[float]]:
    r"""!
    Compute velocity limits in box format:

    \f[
    v_{min} \leq v \leq v_{max}
    \f]

    \param model Pinocchio model.
    \return Velocity limits, with -infinity and +infinity where there is no
    limit.
    """
    no_velocity_limit = np.logical_or(
        model.velocityLimit > 1e20,
        model.velocityLimit < 1e-10,
    )
    v_max = model.velocityLimit.copy()
    v_max[no_velocity_limit] = np.inf
    return v_max


def box_torque_limits(
    model: pin.Model,
) -> Tuple[NDArray[float], NDArray[float]]:
    r"""!
    Compute velocity limits in box format:

    \f[
    \tau_{min} \leq \tau \leq \tau_{max}
    \f]

    \param model Pinocchio model.
    \return Torque limits, with -infinity and +infinity where there is no
    limit.
    """
    no_torque_limit = np.logical_or(
        model.effortLimit > 1e20,
        model.effortLimit < 1e-10,
    )
    tau_max = model.effortLimit.copy()
    tau_max[no_torque_limit] = np.inf
    return tau_max



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

from typing import Tuple

import gymnasium
import numpy as np
from gymnasium import spaces
from numpy.typing import NDArray


class DifferentiateAction(gymnasium.Wrapper):
    def __init__(
        self,
        env,
        min_derivative: NDArray[float],
        max_derivative: NDArray[float],
        action_penalty: float = 0.0,
    ):
        r"""!
        Act on the derivative of the action.

        \param env Environment to wrap.
        \param min_derivative Lower bound on the derivative of the original
            action.
        \param max_derivative Upper bound on the derivative of the original
            action.
        \param action_penalty Weight for an additional penalty on the
            differential action added to the reward.

        \note We assume the original action lives in a vector space.
        """
        super().__init__(env)
        self.action_space = spaces.Box(
            np.float32(min_derivative),
            np.float32(max_derivative),
            shape=env.action_space.shape,
            dtype=np.float32,
        )
        self._integral = np.zeros(env.action_space.shape)
        self.action_penalty = action_penalty

    def reset(self, **kwargs):
        self._integral = np.zeros(self.action_space.shape)
        return self.env.reset(**kwargs)

    def step(
        self,
        action: NDArray[float],
    ) -> Tuple[NDArray[float], float, bool, bool, dict]:
        self._integral = np.clip(
            self._integral + action * self.env.unwrapped.dt,
            self.env.action_space.low,
            self.env.action_space.high,
        )
        observation, reward, terminated, truncated, info = self.env.step(
            self._integral
        )
        wrapped_reward = reward - self.action_penalty * action.dot(action)
        return observation, wrapped_reward, terminated, truncated, info



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

from typing import Union

import gymnasium
import numpy as np
from gymnasium.spaces import Box
from numpy.typing import NDArray

from upkie.utils.filters import low_pass_filter


class LowPassFilterAction(gymnasium.Wrapper):

    """!
    Apply a low-pass filter to the action of an environment.
    """

    filtered_action: NDArray[float]
    time_constant: float
    time_constant_box: Box

    def __init__(self, env, time_constant: Union[float, Box]):
        r"""!
        Initialize wrapper.

        \param env Environment to wrap.
        \param time_constant Cutoff period in seconds of a low-pass filter
            applied to the action. If a Box is provided, couple of lower and
            upper bounds for the action: a new time constant is sampled
            uniformly at random between these bounds at every reset of the
            environment.
        """
        super().__init__(env)
        time_constant_box = (
            time_constant
            if isinstance(time_constant, Box)
            else Box(low=time_constant - 1e-10, high=time_constant + 1e-10)
        )
        self.filtered_action = np.zeros(env.action_space.shape)
        self.time_constant = time_constant_box.sample()
        self.time_constant_box = time_constant_box

    def reset(self, **kwargs):
        self.filtered_action = np.zeros(self.env.action_space.shape)
        self.time_constant = self.time_constant_box.sample()
        return self.env.reset(**kwargs)

    def step(self, action):
        dt = self.env.unwrapped.dt
        if self.time_constant <= 2.0 * dt:
            # Nyquist–Shannon sampling theorem
            return self.env.step(action)

        self.filtered_action = low_pass_filter(
            self.filtered_action,
            self.time_constant,
            action,
            dt,
        )
        return self.env.step(self.filtered_action)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2023 Inria

import math
from typing import Dict, Optional, Tuple

import numpy as np
from gymnasium import spaces
from numpy.typing import NDArray

from upkie.utils.exceptions import UpkieException
from upkie.utils.filters import low_pass_filter
from upkie.utils.robot_state import RobotState

from .upkie_base_env import UpkieBaseEnv


class UpkieGroundVelocity(UpkieBaseEnv):
    r"""!
    Environment where Upkie is used as a wheeled inverted pendulum.

    The environment id is ``UpkieGroundVelocity-v3``. Model assumptions are
    summarized in the <a
    href="https://scaron.info/robotics/wheeled-inverted-pendulum-model.html">following
    note</a>.

    \note For reinforcement learning with neural networks: the observation
    space and action space are not normalized.

    ### Action space

    The action corresponds to the ground velocity resulting from wheel
    velocities. The action vector is simply:

    <table>
        <tr>
            <td><strong>Index</strong></td>
            <td><strong>Description</strong></td>
            </tr>
        <tr>
            <td>``0``</td>
            <td>Ground velocity in [m] / [s].</td>
        </tr>
    </table>

    Note that, while this action is not normalized, [-1, 1] m/s is a reasonable
    range for ground velocities.

    ### Observation space

    Vectorized observations have the following structure:

    <table>
        <tr>
            <td><strong>Index</strong></td>
            <td><strong>Description</strong></td>
        </tr>
        <tr>
            <td>0</td>
            <td>Pitch angle of the base with respect to the world vertical, in
            radians. This angle is positive when the robot leans forward.</td>
        </tr>
        <tr>
            <td>1</td>
            <td>Position of the average wheel contact point, in meters.</td>
        </tr>
        <tr>
            <td>2</td>
            <td>Body angular velocity of the base frame along its lateral axis,
            in radians per seconds.</td>
        </tr>
        <tr>
            <td>3</td>
            <td>Velocity of the average wheel contact point, in meters per
            seconds.</td>
        </tr>
    </table>

    As with all Upkie environments, full observations from the spine (detailed
    in @ref observations) are also available in the ``info`` dictionary
    returned by the reset and step functions.
    """

    class RewardWeights:
        """Weights of the position and velocity terms in rewards."""

        ## @var position
        ## Weight of the position term.
        position: float

        ## @var velocity
        ## Weight of the velocity term.
        velocity: float

        def __init__(self, position: float = 1.0, velocity: float = 1.0):
            r"""!
            Initialize reward weights.

            \param position Weight of the position term.
            \param velocity Weight of the velocity term.
            """
            self.position = position
            self.velocity = velocity

    ## @var action_space
    ## Action space.
    action_space: spaces.box.Box

    ## @var leg_return_period
    ## Time constant for the legs (hips and knees) to revert to their neutral
    ## configuration.
    leg_return_period: float

    ## @var observation_space
    ## Observation space.
    observation_space: spaces.box.Box

    ## @var reward_weights
    ## Weights of the position and velocity terms in rewards.
    reward_weights: "UpkieGroundVelocity.RewardWeights"

    ## @var version
    ## Environment version number.
    version = 3

    ## @var wheel_radius
    ## Wheel radius in [m].
    wheel_radius: float

    def __init__(
        self,
        fall_pitch: float = 1.0,
        frequency: float = 200.0,
        frequency_checks: bool = True,
        init_state: Optional[RobotState] = None,
        leg_return_period: float = 1.0,
        max_ground_velocity: float = 1.0,
        regulate_frequency: bool = True,
        reward_weights: Optional[RewardWeights] = None,
        shm_name: str = "/vulp",
        spine_config: Optional[dict] = None,
        wheel_radius: float = 0.06,
    ):
        r"""!
        Initialize environment.

        \param fall_pitch Fall detection pitch angle, in radians.
        \param frequency Regulated frequency of the control loop, in Hz.
        \param frequency_checks If `regulate_frequency` is set and this
            parameter is true (default), a warning is issued every time the
            control loop runs slower than the desired `frequency`. Set this
            parameter to false to disable these warnings.
        \param init_state Initial state of the robot, only used in simulation.
        \param leg_return_period Time constant for the legs (hips and knees) to
            revert to their neutral configuration.
        \param max_ground_velocity Maximum commanded ground velocity in m/s.
        \param regulate_frequency Enables loop frequency regulation.
        \param reward_weights Coefficients before each reward term.
        \param shm_name Name of shared-memory file.
        \param spine_config Additional spine configuration overriding the
            defaults from ``//config:spine.yaml``. The combined configuration
            dictionary is sent to the spine at every :func:`reset`.
        \param wheel_radius Wheel radius in [m].
        """
        super().__init__(
            fall_pitch=fall_pitch,
            frequency=frequency,
            frequency_checks=frequency_checks,
            init_state=init_state,
            regulate_frequency=regulate_frequency,
            shm_name=shm_name,
            spine_config=spine_config,
        )

        if self.dt is None:
            raise UpkieException("This environment needs a loop frequency")

        reward_weights: UpkieGroundVelocity.RewardWeights = (
            reward_weights
            if reward_weights is not None
            else UpkieGroundVelocity.RewardWeights()
        )

        # gymnasium.Env: observation_space
        MAX_BASE_PITCH: float = np.pi
        MAX_GROUND_POSITION: float = float("inf")
        MAX_BASE_ANGULAR_VELOCITY: float = 1000.0  # rad/s
        observation_limit = np.array(
            [
                MAX_BASE_PITCH,
                MAX_GROUND_POSITION,
                MAX_BASE_ANGULAR_VELOCITY,
                max_ground_velocity,
            ],
            dtype=float,
        )
        self.observation_space = spaces.Box(
            -observation_limit,
            +observation_limit,
            shape=observation_limit.shape,
            dtype=observation_limit.dtype,
        )

        # gymnasium.Env: action_space
        action_limit = np.array([max_ground_velocity], dtype=float)
        self.action_space = spaces.Box(
            -action_limit,
            +action_limit,
            shape=action_limit.shape,
            dtype=action_limit.dtype,
        )

        self.__leg_servo_action = {
            joint: {
                "position": None,
                "velocity": 0.0,
                "maximum_torque": 10.0,  # qdd100 actuators
            }
            for joint in self.model.leg_joints
        }

        self.leg_return_period = leg_return_period
        self.reward_weights = reward_weights
        self.wheel_radius = wheel_radius

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict] = None,
    ) -> Tuple[NDArray[float], Dict]:
        r"""!
        Resets the environment and get an initial observation.

        \param seed Number used to initialize the environment’s internal random
            number generator.
        \param options Currently unused.
        \return
            - ``observation``: Initial vectorized observation, i.e. an element
              of the environment's ``observation_space``.
            - ``info``: Dictionary with auxiliary diagnostic information. For
              Upkie this is the full observation dictionary sent by the spine.
        """
        return super().reset(seed=seed)

    def parse_first_observation(self, spine_observation: dict) -> None:
        r"""!
        Parse first observation after the spine interface is initialized.

        \param spine_observation First observation.
        """
        for joint in self.model.leg_joints:
            position = spine_observation["servo"][joint]["position"]
            self.__leg_servo_action[joint]["position"] = position

    def get_env_observation(self, spine_observation: dict) -> NDArray[float]:
        r"""!
        Extract environment observation from spine observation dictionary.

        \param spine_observation Spine observation dictionary.
        \return Environment observation vector.
        """
        base_orientation = spine_observation["base_orientation"]
        pitch_base_in_world = base_orientation["pitch"]
        angular_velocity_base_in_base = base_orientation["angular_velocity"]
        ground_position = spine_observation["wheel_odometry"]["position"]
        ground_velocity = spine_observation["wheel_odometry"]["velocity"]

        obs = np.empty(4, dtype=float)
        obs[0] = pitch_base_in_world
        obs[1] = ground_position
        obs[2] = angular_velocity_base_in_base[1]
        obs[3] = ground_velocity
        return obs

    def get_leg_servo_action(self) -> Dict[str, Dict[str, float]]:
        """!
        Get servo actions for both hip and knee joints.

        @returns Servo action dictionary.
        """
        for joint in self.model.leg_joints:
            prev_position = self.__leg_servo_action[joint]["position"]
            new_position = low_pass_filter(
                prev_output=prev_position,
                cutoff_period=self.leg_return_period,
                new_input=0.0,
                dt=self.dt,
            )
            self.__leg_servo_action[joint]["position"] = new_position
        return self.__leg_servo_action.copy()

    def get_spine_action(self, action: NDArray[float]) -> dict:
        r"""!
        Convert environment action to a spine action dictionary.

        \param action Environment action.
        \return Spine action dictionary.
        """
        ground_velocity = action[0]
        wheel_velocity = ground_velocity / self.wheel_radius
        servo_dict = self.get_leg_servo_action()
        servo_dict.update(
            {
                "left_wheel": {
                    "position": math.nan,
                    "velocity": +wheel_velocity,
                    "maximum_torque": 1.0,  # mj5208 actuator
                },
                "right_wheel": {
                    "position": math.nan,
                    "velocity": -wheel_velocity,
                    "maximum_torque": 1.0,  # mj5208 actuator
                },
            }
        )
        spine_action = {"servo": servo_dict}
        return spine_action

    def get_reward(
        self,
        observation: NDArray[float],
        action: NDArray[float],
    ) -> float:
        r"""!
        Get reward from observation and action.

        \param observation Environment observation vector.
        \param action Environment action vector.
        \return Reward.
        """
        pitch = observation[0]
        ground_position = observation[1]
        angular_velocity = observation[2]
        ground_velocity = observation[3]

        tip_height = 0.58  # [m]
        tip_position = ground_position + tip_height * np.sin(pitch)
        tip_velocity = (
            ground_velocity + tip_height * angular_velocity * np.cos(pitch)
        )

        std_position = 0.05  # [m]
        position_reward = np.exp(-((tip_position / std_position) ** 2))
        velocity_penalty = -abs(tip_velocity)

        return (
            self.reward_weights.position * position_reward
            + self.reward_weights.velocity * velocity_penalty
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

import abc
from typing import Any, Optional, Tuple

import gymnasium
import numpy as np
from loop_rate_limiters import RateLimiter
from numpy.typing import NDArray
from vulp.spine import SpineInterface

import upkie.config
from upkie.model import Model
from upkie.utils.exceptions import UpkieException
from upkie.utils.nested_update import nested_update
from upkie.utils.robot_state import RobotState


class UpkieBaseEnv(abc.ABC, gymnasium.Env):
    r"""!
    Base class for Upkie environments.

    \note This environment is made to run on a single CPU thread rather than on
    GPU/TPU. The downside for reinforcement learning is that computations are
    not massively parallel. The upside is that it simplifies deployment to the
    real robot, as it relies on the same spine interface that runs on Upkie.
    """

    __frequency: Optional[float]
    __extras: dict
    __rate: Optional[RateLimiter]
    __regulate_frequency: bool
    _spine: SpineInterface
    _spine_config: dict

    ## @var fall_pitch
    ## Fall pitch angle, in radians.
    fall_pitch: float

    ## @var init_state
    ## Initial state for the floating base of the robot, which may be
    ## randomized upon resets.
    init_state: RobotState

    ## @var model
    ## Model parameters for the Upkie wheeled biped.
    model: Model

    def __init__(
        self,
        fall_pitch: float = 1.0,
        frequency: Optional[float] = 200.0,
        frequency_checks: bool = True,
        init_state: Optional[RobotState] = None,
        model: Optional[Model] = None,
        regulate_frequency: bool = True,
        shm_name: str = "/vulp",
        spine_config: Optional[dict] = None,
        spine_retries: int = 10,
    ) -> None:
        r"""!
        Initialize environment.

        \param fall_pitch Fall detection pitch angle, in radians.
        \param frequency Regulated frequency of the control loop, in Hz. Can be
            set even when `regulate_frequency` is false, as some environments
            make use of e.g. `self.dt` internally.
        \param frequency_checks If `regulate_frequency` is set and this
            parameter is true (default), a warning is issued every time the
            control loop runs slower than the desired `frequency`. Set this
            parameter to false to disable these warnings.
        \param init_state Initial state of the robot, only used in simulation.
        \param model Model parameters for the Upkie wheeled biped.
        \param regulate_frequency Enables loop frequency regulation.
        \param shm_name Name of shared-memory file to exchange with the spine.
        \param spine_config Additional spine configuration overriding the
            defaults from ``//config:spine.yaml``. The combined configuration
            dictionary is sent to the spine at every :func:`reset`.
        \param spine_retries Number of times to try opening the shared-memory
            file to communicate with the spine.
        """
        merged_spine_config = upkie.config.SPINE_CONFIG.copy()
        if spine_config is not None:
            nested_update(merged_spine_config, spine_config)
        if regulate_frequency and frequency is None:
            raise UpkieException(f"{regulate_frequency=} but {frequency=}")
        if init_state is None:
            init_state = RobotState(
                position_base_in_world=np.array([0.0, 0.0, 0.6])
            )
        if model is None:
            model = Model()

        self.__extras = {"bullet": {}, "log": {}}
        self.__frequency = frequency
        self.__frequency_checks = frequency_checks
        self.__rate = None
        self.__regulate_frequency = regulate_frequency
        self._spine = SpineInterface(shm_name, retries=spine_retries)
        self._spine_config = merged_spine_config
        self.fall_pitch = fall_pitch
        self.init_state = init_state
        self.model = model

    def __del__(self):
        """!
        Stop the spine when deleting the environment instance.
        """
        self.close()

    def close(self) -> None:
        """!
        Stop the spine properly.
        """
        self._spine.stop()

    @property
    def dt(self) -> Optional[float]:
        """!
        Regulated period of the control loop in seconds, or ``None`` if there
        is no loop frequency regulation.
        """
        return 1.0 / self.__frequency if self.__frequency is not None else None

    @property
    def frequency(self) -> Optional[float]:
        """!
        Regulated frequency of the control loop in Hz, or ``None`` if there is
        no loop frequency regulation.
        """
        return self.__frequency

    def update_init_rand(self, **kwargs) -> None:
        """!
        Update initial-state randomization.

        Keyword arguments are forwarded as is to @ref
        upkie.utils.robot_state_randomization.RobotStateRandomization.update.
        """
        self.init_state.randomization.update(**kwargs)

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict] = None,
    ) -> Tuple[NDArray[float], dict]:
        r"""!
        Resets the spine and get an initial observation.

        \param seed Number used to initialize the environment’s internal random
            number generator.
        \param options Currently unused.
        \return
            - ``observation``: Initial vectorized observation, i.e. an element
              of the environment's ``observation_space``.
            - ``info``: Dictionary with auxiliary diagnostic information. For
              Upkie this is the full observation dictionary sent by the spine.
        """
        super().reset(seed=seed)
        self._spine.stop()
        self.__reset_rate()
        self.__reset_init_state()
        self._spine.start(self._spine_config)
        self._spine.get_observation()  # might be a pre-reset observation
        spine_observation = self._spine.get_observation()
        self.parse_first_observation(spine_observation)
        observation = self.get_env_observation(spine_observation)
        info = {"spine_observation": spine_observation}
        return observation, info

    def __reset_rate(self):
        if self.__regulate_frequency:
            rate_name = f"{self.__class__.__name__} rate limiter"
            self.__rate = RateLimiter(
                self.__frequency,
                name=rate_name,
                warn=self.__frequency_checks,
            )

    def __reset_init_state(self):
        init_state, np_random = self.init_state, self.np_random
        orientation_matrix = init_state.sample_orientation(np_random)
        qx, qy, qz, qw = orientation_matrix.as_quat()
        orientation_quat = np.array([qw, qx, qy, qz])
        position = init_state.sample_position(np_random)
        linear_velocity = init_state.sample_linear_velocity(np_random)
        omega = init_state.sample_angular_velocity(np_random)

        bullet_config = self._spine_config["bullet"]
        reset = bullet_config["reset"]
        reset["orientation_base_in_world"] = orientation_quat
        reset["position_base_in_world"] = position
        reset["linear_velocity_base_to_world_in_world"] = linear_velocity
        reset["angular_velocity_base_in_base"] = omega

    def step(
        self,
        action: NDArray[float],
    ) -> Tuple[NDArray[float], float, bool, bool, dict]:
        r"""!
        Run one timestep of the environment's dynamics. When the end of the
        episode is reached, you are responsible for calling `reset()` to reset
        the environment's state.

        \param action Action from the agent.
        \return
            - ``observation``: Observation of the environment, i.e. an element
              of its ``observation_space``.
            - ``reward``: Reward returned after taking the action.
            - ``terminated``: Whether the agent reached a terminal state,
              which can be a good or a bad thing. When true, the user needs to
              call :func:`reset()`.
            - ``truncated'': Whether the episode is reaching max number of
              steps. This boolean can signal a premature end of the episode,
              i.e. before a terminal state is reached. When true, the user
              needs to call :func:`reset()`.
            - ``info``: Dictionary with auxiliary diagnostic information. For
              us this is the full observation dictionary coming from the spine.
        """
        if self.__regulate_frequency:
            self.__rate.sleep()  # wait until clock tick to send the action
            self.log("rate", {"slack": self.__rate.slack})

        # Act
        spine_action = self.get_spine_action(action)
        for key in ("bullet", "log"):
            if not self.__extras[key]:
                continue
            spine_action[key] = {}
            spine_action[key].update(self.__extras[key])
            self.__extras[key].clear()
        self._spine.set_action(spine_action)

        # Observe
        spine_observation = self._spine.get_observation()
        observation = self.get_env_observation(spine_observation)
        reward = self.get_reward(observation, action)
        terminated = self.detect_fall(spine_observation)
        truncated = False
        info = {"spine_observation": spine_observation}
        return observation, reward, terminated, truncated, info

    def detect_fall(self, spine_observation: dict) -> bool:
        r"""!
        Detect a fall based on the body-to-world pitch angle.

        \param spine_observation Observation dictionary with an "imu" key.
        \return True if and only if a fall is detected.
        """
        pitch = spine_observation["base_orientation"]["pitch"]
        return abs(pitch) > self.fall_pitch

    def parse_first_observation(self, spine_observation: dict) -> None:
        r"""!
        Parse first observation after the spine interface is initialized.

        \param spine_observation First observation.

        This method is an optional way for environments to record some state
        (abstracted away from the agnet) at reset.
        """

    @abc.abstractmethod
    def get_env_observation(self, spine_observation: dict):
        r"""!
        Extract environment observation from spine observation dictionary.

        \param spine_observation Spine observation dictionary.
        \return Environment observation.
        """

    @abc.abstractmethod
    def get_reward(self, observation, action) -> float:
        r"""!
        Get reward from observation and action.

        \param observation Environment observation.
        \param action Environment action.
        \return Reward.
        """

    @abc.abstractmethod
    def get_spine_action(self, action) -> dict:
        r"""!
        Convert environment action to a spine action dictionary.

        \param action Environment action.
        \return Spine action dictionary.
        """

    def log(self, name: str, entry: Any) -> None:
        r"""!
        Log a new entry to the "log" key of the action dictionary.

        \param name Name of the entry.
        \param entry Dictionary to log along with the actual action.
        """
        if isinstance(entry, dict):
            self.__extras["log"][name] = entry.copy()
        else:  # logging values directly
            self.__extras["log"][name] = entry

    def bullet_extra(self, bullet_action: dict) -> None:
        r"""!
        Prepend for the next step an extra action for the Bullet spine.

        This extra action can be for instance a set of external forces applied
        to some robot bodies.

        \param bullet_action Action dictionary processed by the Bullet spine.
        """
        self.__extras["bullet"] = bullet_action.copy()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria


class Model:
    """!
    Structural parameters of Upkie wheeled bipeds.
    """

    def __init__(self):
        """!
        Initialize model with default parameters.
        """
        self.__leg_joints = [
            f"{side}_{joint}"
            for side in ("left", "right")
            for joint in ("hip", "knee")
        ]

    @property
    def leg_joints(self):
        """!
        Get the list of leg (i.e., non-wheel) joints.
        """
        return self.__leg_joints



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria


class UpkieException(Exception):
    """Base class for exceptions raised by Upkie agents."""


class FallDetected(UpkieException):
    """Exception raised when a fall is detected."""


class ModelError(UpkieException):
    """Error related to the robot model."""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

import os
import sys

from .spdlog import logging

__AGENT_CPUID: int = 3
__MODEL_PATH: str = "/sys/firmware/devicetree/base/model"
__ON_RASPI: bool = False

if os.path.exists(__MODEL_PATH):
    with open(__MODEL_PATH, "r", encoding="utf-8") as fh:
        line = fh.readline()
        __ON_RASPI = line.startswith("Raspberry")


def on_raspi() -> bool:
    """!
    Check whether we are running on a Raspberry Pi.
    """
    return __ON_RASPI


def configure_agent_process() -> None:
    """!
    Configure a process to run as an agent on the Raspberry Pi.
    """
    if os.geteuid() != 0:
        logging.info("Re-running as root to set the CPU affinity")
        args = ["sudo", "-E", sys.executable] + sys.argv + [os.environ]
        os.execlpe("sudo", *args)
    os.sched_setaffinity(0, {__AGENT_CPUID})



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

import gymnasium
import numpy as np
from gymnasium import spaces


class ActionObserverEnv(gymnasium.Env):
    def __init__(self):
        action_space = spaces.Box(0.0, 2.0, shape=(1,))
        self.action_space = action_space
        self.observation_space = action_space
        self.dt = 1e-3

    def step(self, action):
        observation = action
        return observation, 0.0, False, False, {}


class ConstantObservationEnv(gymnasium.Env):
    def __init__(self, constant: float):
        self.action_space = spaces.Box(-1.0, 1.0, shape=(1,))
        self.observation_space = spaces.Box(0.0, 2.0, shape=(1,))
        self.constant = constant

    def step(self, action):
        observation = np.array([self.constant])
        return observation, 0.0, False, False, {}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

from .add_action_to_observation import AddActionToObservation
from .differentiate_action import DifferentiateAction
from .low_pass_filter_action import LowPassFilterAction
from .noisify_action import NoisifyAction
from .noisify_observation import NoisifyObservation

__all__ = [
    "AddActionToObservation",
    "DifferentiateAction",
    "LowPassFilterAction",
    "NoisifyAction",
    "NoisifyObservation",
]
